/*******************************************************************************
**	Cirrus Logic GD546x (Laguna) ramdac driver
********************************************************************************
**
   Copyright (C) 1997 Rodolphe Ortalo
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
**
**	$Log: 546x.c,v $
**	Revision 1.10  1998/11/19 02:43:04  marcus
**	License change
**	
**	Revision 1.9  1998/11/04 13:57:24  marcus
**	Changed prototype for kgi_malloc_small()
**	
**	Revision 1.8  1998/11/01 22:45:59  brian
**	Changed INTERNALERROR to INTERNAL_ERROR and removed INTERNALERROR
**	and INVALID_PARAMS from system.h
**	
**	Revision 1.7  1998/10/25 20:03:27  brian
**	Changed cmd to *cmd in ramdac_check_mode and accel_check_mode, plus
**	spot-cleaned a few places Marcus missed during namespace cleanup.
**	(Will get chipset/clock stuff soon, working around Marcus.)
**	
**	Revision 1.6  1998/10/25 18:04:13  marcus
**	More namespace cleanups
**	
**	Revision 1.5  1998/10/25 02:05:44  marcus
**	Namespace cleanup to make KGI/LibGGI interaction possible
**	
**	Revision 1.4  1998/10/04 03:16:42  brian
**	Added private data hook to KGI_COMMON -- clean and recompile kgicon!
**	
**	Revision 1.3  1998/08/25 12:12:10  marcus
**	*** empty log message ***
**	
**	Revision 1.2  1998/07/29 21:04:48  ortalo
**	Driver of CL-GD546x slightly modified to work on kgicon w/ VESA
**	monitor. Still a little strange display sometimes...
**	
**	Revision 1.1  1998/07/07 07:51:01  taylorj
**	*** empty log message ***
**	
**	Revision 1.7  1998/02/09 15:27:00  grimrath
**	New KGI driver Makefile system.  Include faster dependencies, true linking and
**	a predefined driver constant for checking with #ifdef ... #endif
**	
**	VGA, the Mystique driver and kernel files have been adapted to the
**	text <-> dpp change.
**	
**	Revision 1.6  1998/01/27 23:46:14  ortalo
**	Partial support of 2D bitBLT hardware accelerations
**	(tiling management, drawbox in all modes).
**	
**	Revision 1.5  1997/12/18 22:38:34  becka
**	Fixed up RAMDAC drivers. Small spelling error corrected.
**	
**	Revision 1.4  1997/12/17 21:43:45  ortalo
**	Removed the palette initialization (ggi/palette/ *.inc files
**	have disappeared...).
**	
**	Revision 1.3  1997/12/14 21:56:18  becka
**	uint/sint namespace fixed ... please test ...
**	
**	Revision 1.2  1997/10/12 17:56:32  ortalo
**	Ramdac module okay for some graphic modes (and font setting). Hardware
**	cursor support added (but not tested).
**	
**	Revision 1.1  1997/09/14 19:15:03  ortalo
**	New ramdac module for the 546x chipset. Based on the 542x. get/set_clut
**	functions are provided.
**	
**
********************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Rodolphe_Ortalo
#define DRIVER_NAME	"Cirrus Logic CL-GD546x (Laguna) RamDAC driver"
#define DRIVER_REV	"$Revision: 1.10 $"

#include <kgi/kgi.h>
#include <kgi/module.h>

#include "Cirrus/546x.h"

/* Ugly ? -- ortalo */
extern int CLIO_in8(int);
extern void CLIO_out8(int,int);
extern int CLIO_in16(int);
extern void CLIO_out16(int,int);
extern struct mem_region linFrameBuffer;

/*
**	Ramdac description
*/

struct kgi_ramdac 
ramdac = {	"Cirrus Logic",	       		/* manufacturer		*/
	 	"CL-GD546x (Laguna)",		/* model		*/
 		GM_TEXT16 | GM_1BIT | GM_4BIT |	/* supported modes	*/
		GM_8BIT | GM_15BIT | GM_16BIT | GM_24BIT | GM_32BIT ,
	 	0,			 	/* flags		*/
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/
	#warning RAMDAC max freq is different between 5462 and 5464: update !
		NULL,				/* Private data hook	*/
		 DAC_8BIT_PB, 			/* device type		*/
		  {       0,  170000000 },	/* normal mode		*/
		  {       0,         0 },	/* 2xclk mode		*/

		 {{	  0,  170000000 },	/* 	KGIGT_TEXT16	*/
		  {	  0,         0 },	/* 	KGIGT_TEXT32 	*/
		  {       0,  170000000 },	/* 	KGIGT_1BIT		*/
		  {	  0,  170000000 },	/* 	KGIGT_4BIT		*/
/*		  {       0,  170000000 },*/	/* 	KGIGT_8BIT		*/
		  {       0,  170000000 },	/* 	KGIGT_8BIT		*/
		  {	  0,  170000000 },	/* 	KGIGT_15BIT	*/
		  { 	  0,  80000000 },	/* 	KGIGT_16BIT	*/
		  {       0,  170000000 },	/* 	KGIGT_24BIT	*/
		  {	  0,  170000000 }}	/* 	KGIGT_32BIT	*/
         };

/*
** The following parameters indicate the transformations
** we must do before setting a color in the CLUT.
*/
#define SHIFT 8
#define MAGIC 257

/*
**	driver state info
*/

struct dac_private 
{
	kgi_uint	cur_hotx, cur_hoty;	/* cursor hot spot	*/
	kgi_uint	cur_x, cur_y;		/* current position	*/
	kgi_uint	cur_enabled;		/* enabled		*/
};

#define	CUR_HOTX	private->cur_hotx
#define	CUR_HOTY	private->cur_hoty
#define	CUR_X		private->cur_x
#define	CUR_Y		private->cur_y
#define CUR_ENABLED	private->cur_enabled

#define	GET_PRIVATE	struct dac_private *private = KGIM_PRIVATE(dpy,ramdac)

/*
**	Hardware pointer functions
*/


static void dac_set_pointer_shape(struct kgi_display *dpy, 
	struct kgi_pointer64x64 *ptr)
{
	GET_PRIVATE;
	kgi_uint i,j;
	kgiu8 *addr;
	kgi_uint fb_addr;

	switch (ptr->mode) {
		case pmWindows:
		case pmX11:
		case pmThreeColor:
			NOTICE2("Legal cursor mode - but same !");
			break;
		default:
			INTERNAL_ERROR;
	}

	/*	Setting hot spot
	*/
	CUR_HOTX = ptr->hotx;
	CUR_HOTY = ptr->hoty;
	CLIO_out16(CURSOR_PRESET, ((ptr->hotx << 8) & CURSOR_X_PRESET)
				| ( ptr->hoty & CURSOR_Y_PRESET) );

	/* Storing cursor data at the end of the framebuffer
	*/
	fb_addr = (dpy->card->chipset->ramsize - 1024) & ~0x3FF;
	addr = linFrameBuffer.baseptr + fb_addr;
	CLIO_out16(CURSOR_ADDRESS, (fb_addr >> 8) & CURSOR_ADDRESS_MASK);

	DEBUG2("set_pointer_shape(64x64): hot spot (%i,%i) fb_addr:%d addr:%p",
	       CUR_HOTX, CUR_HOTY, fb_addr, addr);

	/*	set pattern (interleaved 64 bits quad-words)
	*/
	for (i = 0; i < 512; i+=8) {
		for (j = 0; j < 8; j++)
			(*addr++) = ptr->xor_mask[i+j];
		for (j = 0; j < 8; j++)	
			(*addr++) = ~(ptr->and_mask[i+j]);
	}

	/*	set colors
	*/
	CLIO_out8(PALETTE_STATE,
		CLIO_in8(PALETTE_STATE) | ACCESS_CURSOR_COLORS);
	/* cursor color 0 */
	CLIO_out8(DAC_PALETTE_ADDRESS_WRITE, 0x00);
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[0].r >> SHIFT); /* red */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[0].g >> SHIFT); /* green */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[0].b >> SHIFT); /* blue */
	/* border color */
	CLIO_out8(DAC_PALETTE_ADDRESS_WRITE, 0x02);
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[1].r >> SHIFT); /* red */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[1].g >> SHIFT); /* green */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[1].b >> SHIFT); /* blue */
	/* cursor color 1 */
	CLIO_out8(DAC_PALETTE_ADDRESS_WRITE, 0x0F);
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[3].r >> SHIFT); /* red */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[3].g >> SHIFT); /* green */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[3].b >> SHIFT); /* blue */
	/* reset palette state */
	CLIO_out8(PALETTE_STATE,
		CLIO_in8(PALETTE_STATE) & ~ACCESS_CURSOR_COLORS);
}


static void dac_show_pointer(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	GET_PRIVATE;

	if (!(CUR_ENABLED)) {
		CLIO_out16(MISCELLANEOUS_CONTROL,
			CLIO_in16(MISCELLANEOUS_CONTROL) | CURSOR_ENABLE);
		CUR_ENABLED = 1;
	}

	if ((CUR_X != x) || (CUR_Y != y)) {

		CUR_X = x;
		CLIO_out16(CURSOR_X, x & CURSOR_X_MASK);
		CUR_Y = y;
		CLIO_out16(CURSOR_Y, y & CURSOR_Y_MASK);
	}
}

static void dac_hide_pointer(struct kgi_display *dpy)
{
	GET_PRIVATE;

	if (CUR_ENABLED) {
		CLIO_out16(MISCELLANEOUS_CONTROL,
			CLIO_in16(MISCELLANEOUS_CONTROL) & ~CURSOR_ENABLE);
		CUR_ENABLED = 0;
	}

	CUR_X = 0x0FFF;
	CUR_Y = 0x0FFF;
}

/* We have a 8 bits CLUT (therefore, we use it !)
** (We also have a 6 bits CLUT, but well, why use it ?)
*/

static void dac_set_clut(struct kgi_display *dpy, kgi_clut *clut,
			kgi_uint first, kgi_uint cnt)
{
	kgiu16 *data = (kgiu16 *) (clut->data + first);
	ASSERT(cnt);

	DEBUG2("Setting %d entries of a new clut", cnt);

	/* First entry to set adress */
	CLIO_out8(DAC_PALETTE_ADDRESS_WRITE, first);
	/* Now setting all entries */
	while (cnt--) {
	  CLIO_out8(DAC_PIXEL_DATA, (*(data++)) >> SHIFT); /* red */
	  CLIO_out8(DAC_PIXEL_DATA, (*(data++)) >> SHIFT); /* green */
	  CLIO_out8(DAC_PIXEL_DATA, (*(data++)) >> SHIFT); /* blue */
	  /* auto-increment of palette adress done by the chipset */
	}
}

static void dac_get_clut(struct kgi_display *dpy,
			 kgi_clut *clut, kgi_uint first, int cnt)
{
	kgiu16 *data = (kgiu16 *) (clut->data + first);

	DEBUG2("Reading %d entries of the clut from %d", cnt, first);

	/* First entry to read adress */
	CLIO_out8(DAC_PALETTE_ADDRESS_READ, first);
	/* Now reading all entries */
	while (cnt--) {
	  *(data++) = CLIO_in8(DAC_PIXEL_DATA) * MAGIC; /* red */
	  *(data++) = CLIO_in8(DAC_PIXEL_DATA) * MAGIC; /* green */
	  *(data++) = CLIO_in8(DAC_PIXEL_DATA) * MAGIC; /* blue */
	  /* auto-increment of palette adress done by the chipset */
	}
}


/*
**	GGI interface
*/
int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	struct dac_private *private;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	if (!(private = kgi_malloc_small(sizeof(*private)))) {

		NOTICE("Failed to allocate ramdac private data!");
		return -ENOMEM;
	}
	KGIM_PRIVATE(dpy,ramdac) = private;
	memset(private, 0, sizeof(*private)); /* Mandatory (for cur_enabled)*/

	card->ramdac=&ramdac;

	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");
}

int
kgim_ramdac_check_mode (struct kgi_display *dpy,
                        struct kgi_mode *mode, enum kgi_tm_cmd *cmd)
{
  int newclock = mode->dclk;
  
  DEBUG2("ramdac_check_mode: clock is initially %d", newclock);

  if (!((1<<mode->request.graphtype)&ramdac.modes))
   {
  	DEBUG1("couldn't handle graphtype %i",mode->request.graphtype);
 	return -E(RAMDAC,MODE_CHECK);
   }

  switch (*cmd)
   {
     case CMD_PROPOSE:
	mode->lclk.mul = mode->clk.mul = 1;
        mode->lclk.div = mode->clk.div = 1;
	return EOK;

     case CMD_LOWER:
       NOTICE2("Ramdac CMD_LOWER");
	if (   (newclock < ramdac.clock[mode->request.graphtype].min)
	    || (newclock < ramdac.normal.min))
	  {
	    DEBUG2("DCLK too low (%i Hz).", newclock);
	    return -E(RAMDAC,CLK_LOW_LIMIT);
	  }

	if (ramdac.clock[mode->request.graphtype].max < newclock)
    	  newclock = ramdac.clock[mode->request.graphtype].max;
	DEBUG2("Ramdac newclock=%i",newclock);
	mode->dclk   = newclock;
	return EOK;

     case CMD_RAISE:
	if (   (newclock > ramdac.clock[mode->request.graphtype].max)
            || (newclock > ramdac.normal.max))
	  {
	    DEBUG2("DCLK too high (%i Hz).", newclock);
	    return -E(RAMDAC,CLK_HIGH_LIMIT);
	  }

	if (newclock < ramdac.clock[mode->request.graphtype].min)
	  newclock = ramdac.clock[mode->request.graphtype].min;

	if (newclock < ramdac.normal.min)
	  newclock = ramdac.normal.min;

	mode->dclk = newclock;
	return EOK;

     case CMD_CHECK:
        if ((mode->lclk.mul != 1) || (mode->lclk.div != 1)
	    /* || (mode->clk.mul  != 1) */ || (mode->clk.div  != 1))
	  return -E(RAMDAC,MODE_CHECK);

	if (   (mode->dclk < ramdac.normal.min)
	    || (mode->dclk < ramdac.clock[mode->request.graphtype].min))
	  return -E(RAMDAC,CLK_LOW_LIMIT);

	if (   (mode->dclk > ramdac.normal.max)
	    || (mode->dclk > ramdac.clock[mode->request.graphtype].max))
	  return -E(RAMDAC,CLK_HIGH_LIMIT);

	return EOK;

    default:
	INTERNAL_ERROR;
	return -E(RAMDAC,BOGUS);
  }
}

void kgim_ramdac_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
	DEBUG2("ramdac_set_mode: for graphtype=%.2x",mode->request.graphtype);

	/* Disable mask on the CLUT */
	CLIO_out8(DAC_PIXEL_MASK, 0xFF);

	switch (mode->request.graphtype) {
		case KGIGT_1BIT:
		case KGIGT_TEXT16:
		case KGIGT_4BIT:
		case KGIGT_8BIT:
		case KGIGT_15BIT:
		case KGIGT_16BIT:
		case KGIGT_24BIT:
		case KGIGT_32BIT:
			break;
		default:
			INTERNAL_ERROR;
	}

        dpy->set_clut = dac_set_clut;
	#warning pointer does not work in text mode. do emulation -- ortalo
	dpy->set_pointer_shape  = dac_set_pointer_shape;
        dpy->show_pointer = dac_show_pointer;
        dpy->hide_pointer = dac_hide_pointer;
	dpy->undo_pointer = NULL;
}

int kgim_ramdac_command(struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg)
{
	/*	this may set 
	**	-  dac_params
	**	-  hardware cursor (to be done)
	**	-  CLUT (or gamma correction table).
	*/
	kgi_clut clut;

	switch(cmd) { 
		case RAMDAC_GETINFO:
			graph->io_res=&ramdac;
			return EOK;
		case RAMDAC_SETINFO:
			if (suser()) {
				memcpy(&ramdac,graph->io_buf,sizeof(ramdac));
				return EOK;
			} else  return -EPERM;
		case RAMDAC_SETCLUT:
			{  clut.size=256;
			   clut.data=(kgi_color *)(graph->io_buf);
			   dac_set_clut(graph->dev.dpy,&clut,0,256);
			} 
			return EOK;
		case RAMDAC_GETCLUT:
			{  clut.size=256;
			   clut.data=(kgi_color *)(graph->io_buf);
			   graph->io_res=graph->io_buf;
			   dac_get_clut(graph->dev.dpy,&clut,0,256);
			} 
			return EOK;
	}
  	return -E(RAMDAC,NOSUP_ALWAYS_CANT);  /* This driver can't do this. */
}





